"""Provides tools to manipulate and format various time-based objects."""
import datetime
from dataclasses import dataclass
from pydantic import BaseModel, NonNegativeInt, NonNegativeFloat, PositiveInt, validate_call


class SplitTime(BaseModel):
    """Data class for holding split time values.

    :ivar int weeks: The number of weeks.
    :ivar int days: The number of days.
    :ivar int hours: The number of hours.
    :ivar int minutes: The number of minutes.
    :ivar int seconds: The number of seconds.
    :ivar int | float milliseconds: The number of milliseconds.
    """
    weeks: NonNegativeInt
    days: NonNegativeInt
    hours: NonNegativeInt
    minutes: NonNegativeInt
    seconds: NonNegativeInt
    milliseconds: NonNegativeInt | NonNegativeFloat


@validate_call
def split_seconds(seconds_in: NonNegativeInt | NonNegativeFloat) -> SplitTime:
    """Split seconds into days, hours, minutes, and seconds, and return a SplitTime object with those values.

    :param int seconds_in: The total number of seconds to split up.

    :return: A SplitTime object with the split weeks, days, hours, minutes, seconds, and milliseconds.
    :rtype: SplitTime
    """
    weeks, remainder = divmod(seconds_in, (60 ** 2) * 24 * 7)
    days, remainder = divmod(remainder, (60 ** 2) * 24)
    hours, remainder = divmod(remainder, 60 ** 2)
    minutes, remainder = divmod(remainder, 60)
    seconds = int(remainder)
    milliseconds = (remainder % 1) * 1000

    return SplitTime(
        weeks=weeks,
        days=days,
        hours=hours,
        minutes=minutes,
        seconds=seconds,
        milliseconds=milliseconds
    )


@validate_call
def day_of_month_suffix(day: PositiveInt) -> str:
    """Return the appropriate suffix for a specified day of the month.

    :param int day: The day of the month to get the suffix for.

    :return: A string that is either 'st', 'nd', 'rd', or 'th'.
    :rtype: str
    """
    if day % 100 in [11, 12, 13]:
        return "th"

    match day % 10:
        case 1:
            return "st"
        case 2:
            return "nd"
        case 3:
            return "rd"
        case _:
            return "th"


@validate_call
def day_of_month_string(day: PositiveInt) -> str:
    """Convert a numerical day of the month to a user-friendly string.

    :param int day: The day of the month to convert.

    :return: A string with the day of the month and an appropriate suffix ('st', 'nd', 'rd', or 'th').
    :rtype: str
    """
    return f"{day}{day_of_month_suffix(day)}"


@validate_call
def timezone_name(datetime_in: datetime.datetime) -> str:
    """Get the full timezone name from a datetime.datetime object.

    :param datetime.datetime datetime_in: The datetime.datetime object to get the timezone for.

    :return: A human-readable string with the full timezone name spelled out.
    :rtype: str
    """
    timezone = datetime_in.astimezone()

    return timezone.tzinfo.tzname(timezone)


@dataclass
class TimeString:
    """Data class with methods to format various datetime objects into human-readable strings."""
    @dataclass
    class TimeDelta:
        """Data class with methods to format datetime.timedelta objects and return human-readable strings."""
        @staticmethod
        @validate_call
        def short(timedelta_in: datetime.timedelta) -> str:
            """Return the datetime.timedelta object as a short human-readable string.

            :param datetime.timedelta timedelta_in: The timedelta to convert.

            :return: The datetime.timedelta object as a short human-readable string
            :rtype: str
            """
            ago_string = ""
            if timedelta_in < datetime.timedelta(0):
                timedelta_in = -timedelta_in
                ago_string += " ago"

            time_parts = split_seconds(timedelta_in.total_seconds())

            time_string = (f"{time_parts.hours}"
                           f":{time_parts.minutes:02}"
                           f":{time_parts.seconds:02}"
                           f"{ago_string}")

            if time_parts.days > 0:
                time_string = f"{time_parts.days}D {time_string}"

            if time_parts.weeks > 0:
                time_string = f"{time_parts.weeks}W {time_string}"

            return time_string

        @staticmethod
        @validate_call
        def long(timedelta_in: datetime.timedelta) -> str:
            """Return the datetime.timedelta object as a long human-readable string.

            :param datetime.timedelta timedelta_in: The timedelta to convert.

            :return: The datetime.timedelta object as a long human-readable string
            :rtype: str
            """
            ago_string = ""
            if timedelta_in < datetime.timedelta(0):
                timedelta_in = -timedelta_in
                ago_string += " ago"

            time_parts = split_seconds(timedelta_in.total_seconds())

            time_strings = []

            for value, name in (
                    (time_parts.weeks, "week"),
                    (time_parts.days, "day"),
                    (time_parts.hours, "hour"),
                    (time_parts.minutes, "minute"),
                    (time_parts.seconds, "second")
            ):
                if name == "second":
                    no_time_strings = len(time_strings) == 0
                else:
                    no_time_strings = False

                if value > 0 or no_time_strings:
                    time_strings.append(f"{value} {name}")
                    if value != 1:
                        time_strings[-1] += "s"

            if len(time_strings) == 1:
                time_string = f"{time_strings[0]}{ago_string}"
            elif len(time_strings) == 2:
                time_string = f"{time_strings[0]} and {time_strings[1]}{ago_string}"
            else:
                time_string = ", ".join(time_strings[:-1])
                time_string += f", and {time_strings[-1]}"
                time_string += ago_string

            return time_string

    @dataclass
    class DateTime:
        """Data class with methods to format datetime.datetime objects and return human-readable strings."""
        @staticmethod
        @validate_call
        def short(datetime_in: datetime.datetime) -> str:
            """Return the datetime.datetime object as a short human-readable string.

            :param datetime.datetime datetime_in: The timedelta to convert.

            :return: The datetime.datetime object as a short human-readable string
            :rtype: str
            """
            now = datetime.datetime.now()

            format_string = "%#I:%M:%S %p"

            if datetime_in.day != now.day or datetime_in.year != now.year:
                format_string = f"%Y/%m/%d @ {format_string}"

            return datetime_in.strftime(format_string).strip()

        @staticmethod
        @validate_call
        def long(datetime_in: datetime.datetime) -> str:
            """Return the datetime.datetime object as a long human-readable string.

            :param datetime.datetime datetime_in: The timedelta to convert.

            :return: The datetime.datetime object as a long human-readable string
            :rtype: str
            """
            now = datetime.datetime.now()

            format_string = "%#I:%M:%S %p"

            if datetime_in.day != now.day or datetime_in.year != now.year:
                format_string = f"%A, %B {day_of_month_string(datetime_in.day)}, %Y at {format_string}"

            time_string = datetime_in.strftime(format_string).strip()
            time_string += f" {timezone_name(datetime_in)}"

            return time_string



"""Provides tools for tracking, computing, and formatting time estimates."""
import datetime
from dataclasses import dataclass
from typing import Any, Annotated, Iterator, Sequence
from pydantic import NonNegativeInt, Field, validate_call

from .time import TimeString
from .constants import EtaDefaults

# TODO: Add a `statistics_string()` method that is focused on all stats (incl. time taken), not just progress
# TODO: Add eta_bar() wrapper for eta_calc()
# TODO: Move to `etatime` package, make `etautil` just a wrapper for compatibility


class Eta:
    """Data class to hold, compute, and format ETA state information and time estimate info.

    :param int total_items: The total number of items to process, used in computations.
    :param int item_index: The index of the item about to be processed (0-indexed).
    :param datetime.datetime start_time: The starting time to use for the computation.
    :param current_time: The time to use for the computation, defaults to the current time.
    :param bool verbose: If we should make strings verbosely or not.
    :param int percent_decimals: The number of decimal places to use in the percentage string.
    :param str not_enough_data_string: The string to return when there is not enough data for the desired computation.

    :ivar int total_items: The total number of items to process, used in computations.
    :ivar int item_index: The index of the item about to be processed (0-indexed).
    :ivar datetime.datetime start_time: The starting time to use for the computation.
    :ivar current_time: The time to use for the computation, defaults to the current time.
    :ivar bool verbose: If we should make strings verbosely or not.
    :ivar int percent_decimals: The number of decimal places to use in the percentage string.
    :ivar str not_enough_data_string: The string to return when there is not enough data for the desired computation.

    :ivar datetime.datetime eta: The estimated completion time.
    :ivar datetime.timedelta time_remaining: The time remaining.
    :ivar float percentage: The completion percentage.
    :ivar datetime.timedelta time_taken: The time taken.
    :ivar str eta_string: The estimated completion time as a human-readable sting.
    :ivar str time_remaining_string: The time remaining as a human-readable string.
    :ivar str percentage_string: The completion percentage as a human-readable string.
    :ivar str time_taken_string: The time taken as a human-readable string.

    :raises pydantic.ValidationError: Raised when a parameter is invalid.
    :raises IndexError: Raised when the index is too large.
    """
    @validate_call
    def __init__(
            self,
            total_items: Annotated[NonNegativeInt, Field(gt=1)],
            item_index: NonNegativeInt,
            start_time: datetime.datetime,
            current_time: datetime.datetime = None,
            verbose: bool = EtaDefaults.verbose,
            percent_decimals: NonNegativeInt = EtaDefaults.percent_decimals,
            not_enough_data_string: str = EtaDefaults.not_enough_data_string
    ):
        if current_time is None:
            current_time = datetime.datetime.now()

        self.total_items = total_items
        self.item_index = item_index
        self.start_time = start_time
        self.current_time = current_time
        self.verbose = verbose
        self.percent_decimals = percent_decimals
        self.not_enough_data_string = not_enough_data_string

        self._validate_item_index(self.item_index)

        self.percent_format = f"{{:.{self.percent_decimals}f}}%"

        self.eta = self._eta()
        self.eta_string = self._eta_string()
        self.time_remaining = self._time_remaining()
        self.time_remaining_string = self._time_remaining_string()
        self.percentage = self._percentage()
        self.percentage_string = self._percentage_string()
        self.time_taken = self._time_taken()
        self.time_taken_string = self._time_taken_string()


    def __str__(self) -> str:
        """Returns the string format of this ETA object.

        :return: The user-friendly progress string.
        :rtype: str
        """
        return self.progress_string()

    def __repr__(self) -> str:
        """Returns the string format of this ETA object.

        :return: The user-friendly progress string.
        :rtype: str
        """
        return self.__str__()

    def _validate_item_index(
            self,
            item_index: NonNegativeInt
    ) -> None:
        """Validate that an index is not larger than the total items and raise an IndexError otherwise.

        Index type and positivity are not validated in this private method because pydantic handles it elsewhere.

        :param int item_index: The index to test.

        :raises IndexError: Raised when the index is too large.

        :rtype: None
        """
        if item_index > self.total_items - 1:
            raise IndexError(f"Item index should be less than {self.total_items - 1} (the total items - 1)")

    def _eta(self) -> datetime.datetime | None:
        """Compute the ETA and return it as a datetime.datetime object.

        :return: The ETA as a datetime.datetime object.
        :rtype: datetime.datetime
        """
        time_remaining = self._time_remaining()
        if time_remaining is None:
            return None

        return self.current_time + time_remaining

    def _eta_string(self) -> str:
        """Compute the ETA and return it as a string.

        :return: The ETA as a human-readable string.
        :rtype: str
        """
        if self.eta is None:
            return self.not_enough_data_string

        if self.verbose:
            return TimeString.DateTime.long(self.eta)

        return TimeString.DateTime.short(self.eta)

    def _time_remaining(self) -> datetime.timedelta | None:
        """Compute the time remaining and return it as a datetime.timedelta object.

        :return: The time remaining as a datetime.timedelta object, None if the index is 0.
        :rtype: datetime.timedelta | None
        """
        percent_done = self._percentage()
        if percent_done <= 0:
            return None

        progress_scale = (1 - percent_done) / percent_done

        return self._time_taken() * progress_scale

    def _time_remaining_string(self) -> str:
        """Compute the time remaining and return it as a string.

        :return: The time remaining as a human-readable string.
        :rtype: str
        """
        if self.time_remaining is None:
            return self.not_enough_data_string

        if self.verbose:
            return TimeString.TimeDelta.long(self.time_remaining)

        return TimeString.TimeDelta.short(self.time_remaining)

    def _percentage(self) -> float:
        """Compute the completion percentage and return it as a float.

        :return: The completion percentage as a float in range on 0.0 - 1.0.
        :rtype: float
        """
        return self.item_index / self.total_items

    def _percentage_string(self) -> str:
        """Compute the completion percentage and return it as a string.

        :return: The completion percentage as a human-readable string.
        :rtype: str
        """
        percent_string = self.percent_format.format(self.percentage * 100)
        if self.verbose:
            percent_string += f" ({self.item_index}/{self.total_items})"

        return percent_string

    def _time_taken(self) -> datetime.timedelta:
        """Compute the time taken and return it as a datetime.timedelta object.

        :return: The time taken as a datetime.timedelta object.
        :rtype: datetime.timedelta
        """
        return self.current_time - self.start_time

    def _time_taken_string(self) -> str:
        """Compute the time taken and return it as a string.

        :return: The time taken as a datetime.timedelta object.
        :rtype: str
        """
        if self.verbose:
            return TimeString.TimeDelta.long(self.time_taken)

        return TimeString.TimeDelta.short(self.time_taken)

    @validate_call
    def progress_string(
            self,
            sep: str = EtaDefaults.sep
    ) -> str:
        """Combine the most useful stats into a string focused on conveying progress and return it.

        :param str sep: The string to use as a seperator between fields.

        :raises pydantic.ValidationError: Raised when a parameter is invalid.

        :return: A human-readable string that includes (in order) percentage, time remaining, and ETA.
        :rtype: str
        """
        percent_string = self._percentage_string()

        if self.item_index <= 0:
            return percent_string

        difference_string = self._time_remaining_string()
        eta_string = self._eta_string()
        if self.verbose:
            difference_string = f"Time remaining: {difference_string}"
            eta_string = f"ETA: {eta_string}"

        return sep.join([percent_string, difference_string, eta_string])

    @validate_call
    def complete(
            self,
            current_time: datetime.datetime = None,
    ) -> None:
        """Set the ETA item to completed (100%).

        :param datetime.datetime current_time: The time to use for the computation, defaults to the current time.

        :rtype: None
        """
        if current_time is None:
            current_time = datetime.datetime.now()

        self.item_index = self.total_items
        self.current_time = current_time

        self.eta = current_time
        self.eta_string = self._eta_string()
        self.time_remaining = datetime.timedelta(0)
        self.time_remaining_string = self._time_remaining_string()
        self.percentage = self._percentage()
        self.percentage_string = self._percentage_string()
        self.time_taken = self._time_taken()
        self.time_taken_string = self._time_taken_string()


class EtaCalculator:
    """Tracks, computes, and formats time estimates.

    :param int total_items: The total number of items to process, used in computations.
    :param datetime.datetime start_time: The starting time used in all calculations, defaults to the current time.
    :param bool verbose: If we should make strings verbosely or not.
    :param int percent_decimals: The number of decimal places to use in the percentage string.
    :param str not_enough_data_string: The string to return when there is not enough data for the desired computation.

    :ivar int total_items: The total number of items to process, used in computations.
    :ivar datetime.datetime start_time: The starting time used in all calculations, defaults to the current time.
    :ivar bool verbose: If we should make strings verbosely or not.
    :ivar int percent_decimals: The number of decimal places to use in the percentage string.
    :ivar str not_enough_data_string: The string to return when there is not enough data for the desired computation.

    :raises pydantic.ValidationError: Raised when a parameter is invalid.
    """
    @validate_call
    def __init__(
            self,
            total_items: Annotated[NonNegativeInt, Field(gt=1)],
            start_time: datetime.datetime = None,
            verbose: bool = EtaDefaults.verbose,
            percent_decimals: NonNegativeInt = EtaDefaults.percent_decimals,
            not_enough_data_string: str = EtaDefaults.not_enough_data_string
    ):
        if start_time is None:
            start_time = datetime.datetime.now()

        self.total_items = None
        self.set_total_items(total_items)

        self.start_time = None
        self.set_start_time(start_time)

        self.verbose = None
        self.set_verbose(verbose)

        self.percent_decimals = None
        self.set_percent_decimals(percent_decimals)

        self.not_enough_data_string = None
        self.set_not_enough_data_string(not_enough_data_string)

    def __str__(self) -> str:
        """Return the string format of this ETA calculator object.

        :return: The user-friendly string representing the calculator object.
        :rtype: str
        """
        return (f"ETA calculator for {self.total_items} items, "
                f"start time = {self.start_time}, "
                f"verbose = {self.verbose}, "
                f"percentage decimal places = {self.percent_decimals}")

    def __repr__(self) -> str:
        """Return the string format of this ETA calculator object.

        :return: The user-friendly string representing the calculator object.
        :rtype: str
        """
        return self.__str__()

    @validate_call
    def get_eta(
            self,
            item_index: NonNegativeInt,
            current_time: datetime.datetime = None
    ) -> Eta:
        """Get the current ETA calculation and return it as an Eta object.

        :param int item_index: The index of the item about to be processed (0-indexed).
        :param datetime.datetime current_time: The time to use for the computation, defaults to the current time.

        :raises pydantic.ValidationError: Raised when a parameter is invalid.
        :raises IndexError: Raised when the index is too large.

        :return: The current ETA calculation as an Eta object.
        :rtype: Eta
        """
        if current_time is None:
            current_time = datetime.datetime.now()

        return Eta(
            total_items=self.total_items,
            item_index=item_index,
            start_time=self.start_time,
            current_time=current_time,
            verbose=self.verbose,
            percent_decimals=self.percent_decimals
        )

    @validate_call
    def set_total_items(
            self,
            total_items: Annotated[NonNegativeInt, Field(gt=1)]
    ) -> None:
        """Set the total number of items to process.

        :param int total_items: The total number of items to process, used in computations.

        :raises pydantic.ValidationError: Raised when a parameter is invalid.

        :rtype: None
        """
        self.total_items = total_items

    @validate_call
    def set_start_time(
            self,
            start_time: datetime.datetime = None
    ) -> None:
        """Set the start time.

        :param datetime.datetime start_time: The starting time to use for the computation, defaults to now.

        :raises pydantic.ValidationError: Raised when a parameter is invalid.

        :rtype: None
        """
        if start_time is None:
            start_time = datetime.datetime.now()

        self.start_time = start_time

    @validate_call
    def set_verbose(
            self,
            verbose: bool
    ) -> None:
        """Set the verbosity of the strings.

        :param bool verbose: If we should make strings verbosely or not.

        :raises pydantic.ValidationError: Raised when a parameter is invalid.

        :rtype: None
        """
        self.verbose = verbose

    @validate_call
    def set_percent_decimals(
            self,
            percent_decimals: NonNegativeInt
    ) -> None:
        """Set the number of decimal places to use in the percentage string.

        :param int percent_decimals: The number of decimal places to use in the percentage string.

        :rtype: None
        """
        self.percent_decimals = percent_decimals

    @validate_call
    def set_not_enough_data_string(
            self,
            not_enough_data_string: str
    ) -> None:
        """Set the string to return when there is not enough data for the desired computation.

        :param str not_enough_data_string: The string to return when there is not enough data for the desired computation.

        :rtype: None
        """
        self.not_enough_data_string = not_enough_data_string


@validate_call
def eta_calculator(
        items: Sequence[Any],
        start_time: datetime.datetime = None,
        verbose: bool = EtaDefaults.verbose,
        percent_decimals: NonNegativeInt = EtaDefaults.percent_decimals,
        not_enough_data_string: str = EtaDefaults.not_enough_data_string
) -> Iterator[tuple[Any, Eta]]:
    """A generator that iterates over the items in a sequence and returns the items in addition to an Eta object.

    :param Sequence[Any] items: The sequence to iterate over.
    :param datetime.datetime start_time: The starting time to use for the computation, defaults to now.
    :param bool verbose: If we should make strings verbosely or not.
    :param int percent_decimals: The number of decimal places to use in the percentage string.
    :param str not_enough_data_string: The string to return when there is not enough data for the desired computation.

    :return: Yield a tuple of the current item and the computed Eta object.
    :rtype: Iterator[tuple[Any, Eta]]
    """
    if start_time is None:
        start_time = datetime.datetime.now()

    calculator = EtaCalculator(
        total_items=len(items),
        start_time=start_time,
        verbose=verbose,
        percent_decimals=percent_decimals,
        not_enough_data_string=not_enough_data_string
    )

    for i, item in enumerate(items):
        yield item, calculator.get_eta(i)



"""Contains constants used in by other components of the module."""
from typing import ClassVar
from dataclasses import dataclass
from pydantic import NonNegativeInt


@dataclass
class EtaDefaults:
    """The defaults to use for the `eta` submodule.

    :cvar bool verbose: If we should make strings verbosely or not.
    :cvar int percent_decimals: The number of decimal places to use in the percentage string.
    :cvar str not_enough_data_string: The string to return when there is not enough data for the desired computation.
    :cvar str sep: The string to use as a seperator between fields.
    """
    verbose: ClassVar[bool] = False
    percent_decimals: ClassVar[NonNegativeInt] = 2
    not_enough_data_string: ClassVar[str] = "not enough data"
    sep: ClassVar[str] = " | "
